unit glider;

interface

uses
    Windows, Direct3D9, D3DX9, dxerr9, global;


const
    brakeForce = 0.009;
    gravity = 0.001;
    gliderLength = 10;
    gliderWidth = 10;
    rotateFactor = 0.015;
    rotateTreshold = -0.2;

type
    CGlider = class
        public
            position: TD3DXVector3;
            rotation: TD3DXVector3;
            velocity: TD3DXVector3;
            userRotationX: single;
        private
            acceleration: TD3DXVector3;
            accelerationForward, accelerationBackward, accelerationLeft, accelerationRight: boolean;
            maxVelocity, accelerateBy: single;
            velocityForward: single;
            velocitySideways: single;
            heightOverGround: single;
            targetRotateX, targetRotateZ: single;
            internalRotationX: single;
        public
            constructor create(heightOverGround, velocity, acceleration: Single);
            procedure setPosition(x, y: single); virtual;
            procedure setRotation(x, y: single); virtual;
            procedure accelerateForward; virtual;
            procedure accelerateBackward; virtual;
            procedure accelerateLeft; virtual;
            procedure accelerateRight; virtual;
            procedure calculateNextFrame; virtual;
    end;

implementation

constructor CGlider.create;
begin
    maxVelocity := velocity;
    accelerateBy := acceleration;
    self.heightOverGround := heightOverGround;

    position.x := 0;
    position.y := 0;
    position.z := 0;

    rotation.x := 0;
    rotation.y := 0;
    rotation.z := 0;

    self.velocity.x := 0;
    self.velocity.y := 0;
    self.velocity.z := 0;

    self.acceleration.x := 0;
    self.acceleration.y := 0;
    self.acceleration.z := 0;

    accelerationForward := false;
    accelerationBackward := false;
    accelerationLeft := false;
    accelerationRight := false;

    velocityForward := 0;
    velocitySideways := 0;

    targetRotateX := 0;
    targetRotateZ := 0;

    userRotationX := 0;
end;

procedure CGlider.setPosition(x: Single; y: Single);
begin
    position.x := x;
    position.y := y;
end;

procedure CGlider.setRotation(x: Single; y: Single);
begin
    userRotationX := x;
    if (userRotationX < -(pi/2)) then userRotationX := -(pi/2);
    if (userRotationX > (pi/2)) then userRotationX := (pi/2);
    rotation.y := y;
end;

procedure CGlider.accelerateForward;
begin
    accelerationForward := true;
end;

procedure CGlider.accelerateBackward;
begin
    accelerationBackward := true;
end;

procedure CGlider.accelerateLeft;
begin
    accelerationLeft := true;
end;

procedure CGlider.accelerateRight;
begin
    accelerationRight := true;
end;

procedure CGlider.calculateNextFrame;
var
    speed:single;
    heightAtXZ: single;
    frontPoint, rearPoint, leftPoint, rightPoint: TD3DXVector3;
    deltaX, deltaY: single;
    alpha: single;
    delta: single;
begin
    speed := accelerateBy * speedFactor;

    if (accelerationForward AND accelerationBackward) then begin
        accelerationForward := false;
        accelerationBackward := false;
    end;

    if (accelerationLeft AND accelerationRight) then begin
        accelerationLeft := false;
        accelerationRight := false;
    end;

    // Beschleunigung:

    acceleration.x := 0;
    acceleration.z := 0;

    // Nach vorne
    if (accelerationForward AND NOT (accelerationLeft OR accelerationRight)) then begin

        acceleration.x := (cos(rotation.y + (pi/2)) * speed);
        acceleration.z := (sin(rotation.y + (pi/2)) * speed);


    // Vorne-Links
    end else if (accelerationForward AND accelerationLeft) then begin

        acceleration.x := (cos(rotation.y + (pi/2) + (pi/4)) * speed);
        acceleration.z := (sin(rotation.y + (pi/2) + (pi/4)) * speed);


    // Vorne-Rechts
    end else if (accelerationForward AND accelerationRight) then begin

        acceleration.x := (cos(rotation.y + (pi/2) - (pi/4)) * speed);
        acceleration.z := (sin(rotation.y + (pi/2) - (pi/4)) * speed);

    // Nach hinten
    end else if (accelerationBackward AND NOT (accelerationLeft OR accelerationRight)) then begin

        acceleration.x := -(cos(rotation.y + (pi/2)) * speed);
        acceleration.z := -(sin(rotation.y + (pi/2)) * speed);

    // Hinten-links
    end else if (accelerationBackward AND accelerationLeft) then begin

        acceleration.x := (cos(rotation.y - (pi/2) - (pi/4)) * speed);
        acceleration.z := (sin(rotation.y - (pi/2) - (pi/4)) * speed);

    // Hinten-rechts
    end else if (accelerationBackward AND accelerationRight) then begin

        acceleration.x := (cos(rotation.y - (pi/2) + (pi/4)) * speed);
        acceleration.z := (sin(rotation.y - (pi/2) + (pi/4)) * speed);

    // Links
    end else if ((NOT (accelerationForward or accelerationBackward)) AND accelerationLeft) then begin

        acceleration.x := (cos(rotation.y + pi) * speed);
        acceleration.z := (sin(rotation.y + pi) * speed);

    // Rechts
    end else if ((NOT (accelerationForward or accelerationBackward)) AND accelerationRight) then begin

        acceleration.x := -(cos(rotation.y + pi) * speed);
        acceleration.z := -(sin(rotation.y + pi) * speed);

    end;

    velocity.x := velocity.x - (brakeForce * speedFactor * velocity.x);
    velocity.z := velocity.z - (brakeForce * speedFactor * velocity.z);

    velocity.x := velocity.x - acceleration.x;
    velocity.z := velocity.z - acceleration.z;

    if (velocity.x > maxVelocity) then velocity.x := maxVelocity;
    if (velocity.x < -maxVelocity) then velocity.x := -maxVelocity;

    if (velocity.z > maxVelocity) then velocity.z := maxVelocity;
    if (velocity.z < -maxVelocity) then velocity.z := -maxVelocity;


    // Position berechnen
    position.x := position.x + velocity.x * speedFactor;
    position.z := position.z + velocity.z * speedFactor;
    heightAtXZ := terrain.getHeightAtPos(position.x, position.z) - heightOverGround;


    // Drehung
    if ((position.y - rotateTreshold) > heightAtXZ) then begin


        // X-Achse
        frontPoint.x := position.x + (cos(rotation.y + (pi/2)) * gliderLength);
        frontPoint.z := position.z + (sin(rotation.y + (pi/2)) * gliderLength);
        frontPoint.y := terrain.getHeightAtPos(frontPoint.x, frontPoint.z);

        rearPoint.x := position.x - (cos(rotation.y + (pi/2)) * gliderLength);
        rearPoint.z := position.z - (sin(rotation.y + (pi/2)) * gliderLength);
        rearPoint.y := terrain.getHeightAtPos(rearPoint.x, rearPoint.z);

        deltaY := frontPoint.y - rearPoint.y;
        deltaX := 2 * gliderLength;
        alpha := arctan(deltaY / deltaX);

        targetRotateX := alpha;

        // Z-Achse
        leftPoint.x := position.x + (cos(rotation.y + pi) * gliderWidth);
        leftPoint.z := position.z + (sin(rotation.y + pi) * gliderWidth);
        leftPoint.y := terrain.getHeightAtPos(leftPoint.x, leftPoint.z);

        rightPoint.x := position.x - (cos(rotation.y + pi) * gliderWidth);
        rightPoint.z := position.z - (sin(rotation.y + pi) * gliderWidth);
        rightPoint.y := terrain.getHeightAtPos(rightPoint.x, rightPoint.z);

        deltaY := rightPoint.y - leftPoint.y;
        deltaX := 2 * gliderWidth;
        alpha := arctan(deltaY / deltaX);

        targetRotateZ := -alpha;

    end else begin

        targetRotateX := -pi/6;
        targetRotateZ := 0;

    end;


    // Hhe
    if (position.y < heightAtXZ) then begin
        // freier Fall
        acceleration.y := acceleration.y + gravity * speedfactor;
        velocity.y := velocity.y + acceleration.y;
        position.y := position.y + velocity.y * speedFactor;
        if position.y > heightAtXZ then position.y := heightAtXZ;
    end else begin
        // Bodenkontakt
        acceleration.y := 0;
        velocity.y := 0;
        position.y := heightAtXZ;

    end;

    // Drehung - Dmpfung
    delta := abs(rotation.x - targetRotateX);
    if (internalRotationX > targetRotateX) then begin
       internalRotationX := internalRotationX - (rotateFactor * delta * speedFactor);
       if (internalRotationX < targetRotateX) then internalRotationX := targetRotateX;
    end;
    if (internalRotationX < targetRotateX) then begin
       internalRotationX := internalRotationX + (rotateFactor * delta * speedFactor);
       if (internalRotationX > targetRotateX) then internalRotationX := targetRotateX;
    end;

    delta := abs(rotation.z - targetRotateZ);
    if (rotation.z > targetRotateZ) then begin
       rotation.z := rotation.z - (rotateFactor * delta * speedFactor);
       if (rotation.z < targetRotateZ) then rotation.z := targetRotateZ;
    end;
    if (rotation.z < targetRotateZ) then begin
       rotation.z := rotation.z + (rotateFactor * delta * speedFactor);
       if (rotation.z > targetRotateZ) then rotation.z := targetRotateZ;
    end;

    rotation.x := internalRotationX + userRotationX;

    // Fr den nchsten Frame wieder zurcksetzen
    accelerationForward := false;
    accelerationBackward := false;
    accelerationLeft := false;
    accelerationRight := false;

end;

end.