(*----------------------------------------------------------------------------*
 *                                                                            *
 *         M A R S                                                            *
 *                                                                            *
 *         by Jan Strter                                                     *
 *                                                                            *
 *----------------------------------------------------------------------------*)

 unit landscape;

interface

uses
  Windows, Messages, Direct3D9, D3DX9, dxerr9,
  D3DUtil, DXUtil, sysutils, math, directinput;

const
  D3DFVF_CUSTOMVERTEX = D3DFVF_XYZ or D3DFVF_TEX1 or D3DFVF_TEX2 or D3DFVF_TEX3 or D3DFVF_NORMAL;
  D3DFVF_TEMP = D3DFVF_XYZ or D3DFVF_NORMAL;
  mapSize = 512;
  treeDepth = 9;
  cellWidth = 4;
  heightFactor = 0.6;
  STD_TEXTUREFACTOR = 40;

type
  PCustomVertex = ^TCustomVertex;
  TCustomVertex = packed record
    position: TD3DXVector3;
    normal: TD3DVector;
    tu0, tv0: Single;
    tu1, tv1: Single;
    tu2, tv2: Single;
  end;

  PCustomVertexArray = ^TCustomVertexArray;
  TCustomVertexArray = array [0..8] of TCustomVertex;

  PCustomTempVertex = ^TCustomTempVertex;
  TCustomTempVertex = packed record
    position: TD3DXVector3;
    normal: TD3DVector;
  end;

  PCustomTempVertexArray = ^TCustomTempVertexArray;
  TCustomTempVertexArray = array [0..8] of TCustomTempVertex;

  PQuadTree = ^TQuadTree;
  TQuadTree = record
     parent: PQuadTree;
     hasChildren: boolean;
     children: array[0..3] of PQuadTree;
     vertexData: packed array[0..8] of TCustomVertex;
     level: integer;
     size: integer;
     left, top: integer;
     center: TD3DXVector3;
     boundingBox_min, boundingBox_max: TD3DXVector3;
     quadrant: integer;
     neighbour: packed array[0..3] of PQuadTree;
     render: boolean;
     texture: IDirect3DTexture9;
  end;


  CLandscape = class

    // Terrain-Daten
    textureFactor:single;
    vertexIndices: packed array[0..23] of word;
    tempVertexBuffer: IDirect3DVertexBuffer9;
    normals:packed array[0..mapSize, 0..mapSize] of packed record
       n1: TD3DVector;
       n2: TD3DVector;
       average: TD3DVector;
    end;
    heights:packed array[0..mapSize, 0..mapSize] of single;
    textures:packed array[0..mapSize, 0..mapSize] of byte;
    texture_coords:packed array[0..mapSize, 0..mapSize] of packed record
        tu, tv: Single;
    end;

    // Quad-Tree
    tree: PQuadTree;
    neighbours: packed array[0..mapSize, 0..mapSize, 0..treeDepth] of record
                tree: pointer;
                quadrant: integer;
    end;

    m_pd3dDevice: IDirect3DDevice9;

  public
    function Render: HRESULT;
    function InitGeometry: HRESULT;
    function createQuadTree(left, top, level, size: integer; parent: PQuadTree; quadrant: integer): PQuadTree;
    procedure destroyQuadTree(tree: PQuadTree);
    procedure queryRenderQuadTree(tree: PQuadTree);
    procedure renderQuadTree(tree: PQuadTree; testChildren: boolean);
    procedure fillNeighbours(tree: PQuadTree);
    procedure assignNeighbours(tree: PQuadTree);
    function getHeightAtPos(x, y: single):single;
    procedure createLightMap;
    procedure updateDevice(device: IDirect3DDevice9);
    constructor Create(device: IDirect3DDevice9);
    destructor Destroy;
  end;

///////////////////////////////////////////////////////////////////////////
implementation
///////////////////////////////////////////////////////////////////////////

uses
    global;

constructor CLandscape.Create;
begin
     updateDevice(device);
end;

destructor CLandscape.Destroy;
begin
    destroyQuadTree(tree);
end;

procedure CLandscape.updateDevice;
begin
     m_pd3dDevice := device;
end;

//-----------------------------------------------------------------------------
// Name: CLandscape()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CLandscape.Render: HRESULT;
var
   bla:PAnsiChar;
   matPos: TD3DXMatrix;
   i: integer;
   mapx, mapy: integer;
begin
    m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, dWord(true));
    m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, dWord(true));
    D3DXMatrixIdentity(matPos);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matPos);

    // Set up the default texture states.
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLORARG1, D3DTA_CURRENT);
    m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLORARG2, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    // Set up the default sampler states.
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP );
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP );

    m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSU,  D3DTADDRESS_WRAP );
    m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSV,  D3DTADDRESS_WRAP );

    m_pd3dDevice.SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP );
    m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP );

    // Render the vertex buffer contents

    m_pd3dDevice.SetTexture(0, overall_textures[0]);
    m_pd3dDevice.SetTexture(1, detail_texture);
    m_pd3dDevice.SetTexture(2, lightmap);

    m_pd3dDevice.SetFVF(D3DFVF_CUSTOMVERTEX);
    queryRenderQuadTree(tree);
    renderQuadTree(tree, true);

  Result:= S_OK;
end;


function computeNormal(vec: array of TCustomTempVertex): TD3DVector;
var
   normal: TD3DVector;
   vec1, vec2: TD3DVector;
begin
     D3DXVec3Subtract(vec1, vec[1].position, vec[0].position);
     D3DXVec3Subtract(vec2, vec[2].position, vec[0].position);
     D3DXVec3Cross(normal, vec1, vec2);
     D3DXVec3Normalize(normal, normal);
     normal.y := -normal.y;
     result := normal;
end;

 function CLandscape.getHeightAtPos;
var
   mapX, mapY: single;
   height:single;

   floor_x, floor_y, ceil_x, ceil_y: integer;
   fraction_x, fraction_y, one_minus_x, one_minus_y, h1, h2, h3, h4, v1, v2: single;
begin
     mapX := (-x - x_Versatz) / cellWidth;
     mapY := (-y - z_versatz) / cellWidth;

     result := -100;
     if (trunc(mapX) >= 1) and (trunc(mapX) < mapSize - 1) then begin
        if (trunc(mapY) >= 1) and (trunc(mapY) < mapSize - 1) then begin

            floor_x := floor(mapX);
            floor_y := floor(mapY);
            ceil_x := floor_x + 1;
            ceil_y := floor_y + 1;
            fraction_x := mapX - floor_x;
            fraction_y := mapY - floor_y;
            one_minus_x := 1.0 - fraction_x;
            one_minus_y := 1.0 - fraction_y;

            h1 := heights[floor_x, floor_y];
            h2 := heights[ceil_x, floor_y];
            h3 := heights[floor_x, ceil_y];
            h4 := heights[ceil_x, ceil_y];

            v1 := (one_minus_x * h1 + fraction_x * h2);
            v2 := (one_minus_x * h3 + fraction_x * h4);
            height := (one_minus_y * v1) + (fraction_y * v2);

             height := height - y_versatz;

             result := height;
        end;
     end;
end;


//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Creates the scene geometry
//-----------------------------------------------------------------------------
function CLandscape.InitGeometry: HRESULT;

type
    TColor = packed record
           r:byte;
           g:byte;
           b:byte;
           a:byte;
    end;

    TUV = packed record
        u, v: single;
    end;

var
  pVertices: PCustomTempVertexArray;
  x, y, a, b: integer;
  temp, temp2, temp3, temp4:single;
  lockedRect, overallRect: TD3DLOCKEDRECT;
  pixel:TColor;
  y_ofs, x_ofs, index, i:integer;
  zeiger: pointer;
  the_normal:TD3DVector;
  tmpNormal: TD3DVector;
  tmpUV, tmpUV2, tmpUV3, tmpUV4: TUV;
  length_u, length_v:single;
  tmpVec:TD3DXVECTOR3;
  flipVertecies:array[0..2] of TCustomTempVertex;
  lightsource: TD3DXVector3;
begin
  Result:= E_FAIL;


  // Create the vertex buffer.
  if FAILED(m_pd3dDevice.CreateVertexBuffer(6*mapSize*mapSize*SizeOf(TCustomVertex),
                                            0, D3DFVF_TEMP,
                                            D3DPOOL_MANAGED, tempVertexBuffer, nil))
  then Exit;

  if FAILED(tempVertexBuffer.Lock(0, 0, Pointer(pVertices), 0))
  then Exit;

  x_versatz := -mapSize * cellWidth / 2;
  y_versatz := -255 * heightFactor;
  z_versatz := -mapSize * cellWidth / 2;
  textureFactor := STD_TEXTUREFACTOR;



  // Hhen berechnen
  if FAILED(heightmap.LockRect(0, lockedRect, nil, 0)) then exit;
  y_ofs := 0;
  for y := 0 to mapSize-1 do begin

      x_ofs := 0;

      for x := 0 to mapSize-1 do begin

          // Adresse zum Auslesen der Textur berechnen
          zeiger := ptr(integer(lockedRect.pBits) + y_ofs + x_ofs);
          move(zeiger^, pixel, SizeOf(TColor));
          heights[x, y] :=  -pixel.r * heightFactor;

          inc(x_ofs, SizeOf(TColor));

      end;
      heights[mapSize, y] := heights[mapSize - 1, y];
      inc(y_ofs, lockedRect.Pitch);

  end;
  for x := 0 to mapSize do begin
      heights[x, mapSize] := heights[x, mapSize - 1];
  end;
  heightmap.UnlockRect(0);


  // Netz erstellen (inkl. Flchen-Normale fr jedes Dreieck)
  for y := 0 to mapSize-1 do begin
      for x := 0 to mapSize-1 do begin

          // Hhen fr die vier Ecken der Zelle auslesen
          temp := - heights[x, y];
          temp2 := -heights[x + 1, y];
          temp3 := -heights[x + 1, y + 1];
          temp4 := -heights[x, y + 1];

          // Index in den Vertex-Buffer berechnen
          index := (y * mapSize * 6) + (x * 6);

          // 1. Dreieck erstellen
          pVertices[index].position := D3DXVector3(x * cellWidth + x_versatz, temp + y_versatz, y * cellWidth + z_versatz);
          inc(index);
          pVertices[index].position := D3DXVector3(x * cellWidth + x_versatz, temp4 + y_versatz, (y+1) * cellWidth + z_versatz);
          inc(index);
          pVertices[index].position := D3DXVector3((x+1) * cellWidth + x_versatz, temp2 + y_versatz, y * cellWidth + z_versatz);
          inc(index);

          // Normale fr 1. Dreieck berechnen
          flipVertecies[0] := pVertices[index - 3 + 0];
          flipVertecies[1] := pVertices[index - 3 + 1];
          flipVertecies[2] := pVertices[index - 3 + 2];
          the_normal := computeNormal(flipVertecies);
          normals[x, y].n1 := the_normal;
          for i := 0 to 2 do begin
              pVertices[index - 3 + i].normal := the_normal;
          end;

          // 2. Dreieck erstellen
          pVertices[index].position := D3DXVector3(x * cellWidth + x_versatz, temp4 + y_versatz, (y+1) * cellWidth + z_versatz);
          inc(index);
          pVertices[index].position := D3DXVector3((x+1) * cellWidth + x_versatz, temp3 + y_versatz, (y+1) * cellWidth + z_versatz);
          inc(index);
          pVertices[index].position := D3DXVector3((x+1) * cellWidth + x_versatz, temp2 + y_versatz, y * cellWidth + z_versatz);
          inc(index);

          // Normale fr 1. Dreieck berechnen
          flipVertecies[0] := pVertices[index - 3 + 0];
          flipVertecies[1] := pVertices[index - 3 + 1];
          flipVertecies[2] := pVertices[index - 3 + 2];
          the_normal := computeNormal(flipVertecies);
          normals[x, y].n2 := the_normal;
          for i := 0 to 2 do begin
              pVertices[index - 3 + i].normal := the_normal;
          end;

      end;
  end;

  // Normalen fr Gouraud-Shading berechnen (Mittelwert der Flchennormalen an einem Scheitelpunkt)
  for y := 1 to mapSize-2 do begin
      for x := 1 to mapSize-2 do begin
          index := (y * mapSize * 6) + (x * 6);

          tmpNormal.x := tmpNormal.x + normals[x - 1, y].n2.x;
          tmpNormal.x := tmpNormal.x + normals[x, y - 1].n1.x;
          tmpNormal.x := tmpNormal.x + normals[x, y - 1].n2.x;
          tmpNormal.x := tmpNormal.x + normals[x, y].n1.x;
          tmpNormal.x := tmpNormal.x + normals[x - 1, y].n2.x;
          tmpNormal.x := tmpNormal.x + normals[x - 1, y].n1.x;

          tmpNormal.y := tmpNormal.y + normals[x - 1, y].n2.y;
          tmpNormal.y := tmpNormal.y + normals[x, y - 1].n1.y;
          tmpNormal.y := tmpNormal.y + normals[x, y - 1].n2.y;
          tmpNormal.y := tmpNormal.y + normals[x, y].n1.y;
          tmpNormal.y := tmpNormal.y + normals[x - 1, y].n2.y;
          tmpNormal.y := tmpNormal.y + normals[x - 1, y].n1.y;

          tmpNormal.z := tmpNormal.z + normals[x - 1, y].n2.z;
          tmpNormal.z := tmpNormal.z + normals[x, y - 1].n1.z;
          tmpNormal.z := tmpNormal.z + normals[x, y - 1].n2.z;
          tmpNormal.z := tmpNormal.z + normals[x, y].n1.z;
          tmpNormal.z := tmpNormal.z + normals[x - 1, y].n2.z;
          tmpNormal.z := tmpNormal.z + normals[x - 1, y].n1.z;

          tmpNormal.x := tmpNormal.x / 6;
          tmpNormal.y := tmpNormal.y / 6;
          tmpNormal.z := tmpNormal.z / 6;

          D3DXVec3Normalize(tmpNormal, tmpNormal);
          normals[x, y].average := tmpNormal;

      end;
  end;

  // tempVertexBuffer wird nun nicht mehr bentigt, freigeben (ca. 70 MB RAM!)
  tempVertexBuffer.Unlock;
  tempVertexBuffer := NIL;

  // Lightmap erstellen
  createLightMap;

  // Vertex-Buffer fllen
  vertexIndices[0] := 0;
  vertexIndices[1] := 3;
  vertexIndices[2] := 1;
  vertexIndices[3] := 3;
  vertexIndices[4] := 4;
  vertexIndices[5] := 1;
  vertexIndices[6] := 1;
  vertexIndices[7] := 4;
  vertexIndices[8] := 2;
  vertexIndices[9] := 4;
  vertexIndices[10] := 5;
  vertexIndices[11] := 2;
  vertexIndices[12] := 3;
  vertexIndices[13] := 6;
  vertexIndices[14] := 4;
  vertexIndices[15] := 6;
  vertexIndices[16] := 7;
  vertexIndices[17] := 4;
  vertexIndices[18] := 4;
  vertexIndices[19] := 7;
  vertexIndices[20] := 5;
  vertexIndices[21] := 5;
  vertexIndices[22] := 7;
  vertexIndices[23] := 8;

  // Textur-Indezies und Koordinaten fr die "Overall-Texture" erstellen
  i := 0;
  for y := 0 to 3 do begin
      for x := 0 to 3 do begin
          for b := 0 to (MAPSIZE div 4) - 1 do begin
              for a := 0 to (MAPSIZE div 4) - 1 do begin
                  textures[a + ((MAPSIZE div 4) * x), b + ((MAPSIZE div 4) * y)] := i;
                  texture_coords[a + ((MAPSIZE div 4) * x), b + ((MAPSIZE div 4) * y)].tu := a / 128;
                  texture_coords[a + ((MAPSIZE div 4) * x), b + ((MAPSIZE div 4) * y)].tv := b / 128;
              end;
          end;
          inc(i);
      end;
  end;

  // Quad-Tree mit den Terrain-Daten erstellen
  tree := createQuadTree(0, 0, 0, mapSize div 2, NIL, 0);
  fillNeighbours(tree);
  assignNeighbours(tree);

  Result:= S_OK;
end;


function CLandscape.createQuadTree;
type
    TColor = packed record
           r:byte;
           g:byte;
           b:byte;
           a:byte;
    end;

    TUV = packed record
        u, v: single;
    end;

var
   pVertices: PCustomVertexArray;
   quadTree: PQuadTree;
   i, a, b, index: integer;
   src_x, src_y: integer;
   x, y: integer;
   max_height, min_height: single;
begin
     if size < 1 then begin
        result := NIL;
        exit;
     end;

     New(quadTree);
     result := quadTree;

     quadTree^.parent := parent;
     quadTree^.level := level;
     quadTree^.size := size;
     quadTree^.left := left;
     quadTree^.top := top;
     quadTree^.quadrant := quadrant;
     quadTree^.center := D3DXVector3((left + size) * cellWidth + x_versatz, y_versatz, (top + size) * cellWidth + z_versatz);
     quadTree^.texture := overall_textures[textures[left, top]];
//     quadTree^.texture := overall_textures[0];
     pVertices := @quadTree^.vertexData;
     index := 0;

     for b := 0 to 2 do begin
        for a := 0 to 2 do begin

            src_x := a * size + left;
            src_y := b * size + top;

            // Position des Vertex
            pVertices[index].position := D3DXVector3(src_x * cellWidth + x_versatz, - heights[src_x, src_y] + y_versatz, src_y * cellWidth + z_versatz);

            // Textur-Koordinaten fr "Overall-Texture"
            pVertices[index].tu0 := texture_coords[src_x, src_y].tu;
            if a = 2 then begin
              if ((src_x mod 128) = 0) then pVertices[index].tu0 := 1;
            end;
            pVertices[index].tv0 := texture_coords[src_x, src_y].tv;
            if b = 2 then begin
              if ((src_y mod 128) = 0) then pVertices[index].tv0 := 1;
            end;

            // Textur-Koordinaten fr Detail-Textur
            pVertices[index].tu1 := src_x / mapSize * textureFactor;
            pVertices[index].tv1 := src_y / mapSize * textureFactor;

            // Textur-Koordinaten fr Light-Map
            pVertices[index].tu2 := src_x / mapSize;
            pVertices[index].tv2 := src_y / mapSize;

            // Normale fr den Vertex
            pVertices[index].normal := normals[src_x, src_y].average;
            inc(index);

        end;
    end;

    quadTree^.boundingBox_min := pVertices[0].position;
    quadTree^.boundingBox_max := quadTree^.boundingBox_min;

    for i := 0 to 8 do
    begin
      if pVertices[i].position.x < quadTree^.boundingBox_min.x then quadTree^.boundingBox_min.x := pVertices[i].position.x;
      if pVertices[i].position.z < quadTree^.boundingBox_min.z then quadTree^.boundingBox_min.z := pVertices[i].position.z;

      if pVertices[i].position.x > quadTree^.boundingBox_max.x then quadTree^.boundingBox_max.x := pVertices[i].position.x;
      if pVertices[i].position.z > quadTree^.boundingBox_max.z then quadTree^.boundingBox_max.z := pVertices[i].position.z;
    end;

    for y := top to top + (size * 2) do begin
      for x := left to left + (size * 2) do begin
          if (- heights[x, y] + y_versatz) < quadTree^.boundingBox_min.y then quadTree^.boundingBox_min.y := (- heights[x, y] + y_versatz);
          if (- heights[x, y] + y_versatz) > quadTree^.boundingBox_max.y then quadTree^.boundingBox_max.y := (- heights[x, y] + y_versatz);
      end;
    end;

{    quadTree^.boundingBox_max.y := quadTree^.boundingBox_max.y + 3;
    quadTree^.boundingBox_min.y := quadTree^.boundingBox_min.y - 3;}

    // ggf. Kind-Elemente berechnen
    if size >= 2 then begin
       quadTree^.children[0] := createQuadTree(left, top, level + 1, size div 2, quadTree, 0);
       quadTree^.children[1] := createQuadTree(left + size, top, level + 1, size div 2, quadTree, 1);
       quadTree^.children[2] := createQuadTree(left + size, top + size, level + 1, size div 2, quadTree, 2);
       quadTree^.children[3] := createQuadTree(left, top + size, level + 1, size div 2, quadTree, 3);
       quadTree^.hasChildren := true;
    end else begin
        quadTree^.hasChildren := false;
    end;
end;

procedure CLandscape.destroyQuadTree;
var
   i: integer;
begin
     if tree^.hasChildren then begin
        for i := 0 to 3 do begin
            destroyQuadTree(tree^.children[i]);
        end;
     end;
     dispose(tree);
end;

procedure CLandscape.queryRenderQuadTree;
var
   i: integer;
   vector: TD3DXVector3;
   distance: single;
begin

     // Abstand zum Betrachter berechnen
     vector.x := -viewport.position.x - tree^.center.x;
     vector.y := -viewport.position.y - tree^.center.y;
     vector.z := -viewport.position.z - tree^.center.z;
     distance := D3DXVec3Length(vector);

     // Detail-Stufe berechnen
     tree^.render := false;
     case tree^.level of
{          4: if distance >= 500 then tree^.render := true;
          5: if distance >= 400 then tree^.render := true;
          6: if distance >= 300 then tree^.render := true;
          7: if distance >= 200 then tree^.render := true;}
          2: if distance >= 1200 then tree^.render := true;
          3: if distance >= 900 then tree^.render := true;
          4: if distance >= 750 then tree^.render := true;
          5: if distance >= 500 then tree^.render := true;
          6: if distance >= 350 then tree^.render := true;
          7: if distance >= 200 then tree^.render := true;
          8: tree^.render := true;
     end;

     // Wenn dieses Element nicht gerendert werden soll, Kind-Elemente testen
     if not tree^.render then begin
         if tree^.hasChildren then begin
            for i := 0 to 3 do begin
                queryRenderQuadTree(tree^.children[i]);
            end;
         end;
     end;
end;

procedure CLandscape.renderQuadTree;
var
   i: integer;
   smooth: packed array[0..8] of TCustomVertex;
   smoothHeight: single;
   copied: boolean;
   testRes: integer;
begin
     if testChildren then begin
        testRes := isVisible(tree^.boundingBox_min, tree^.boundingBox_max);
        if testRes = 2 then testChildren := false;
     end else begin
        testRes := 2;
     end;

     testRes := 1;

     if testRes > 0 then begin

         if tree^.render then begin

            copied := false;
            case tree^.quadrant of
                 0: begin
                         if tree^.neighbour[0] <> NIL then begin
                           if tree^.neighbour[0].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[0].vertexData[2].position.y + tree^.neighbour[0].vertexData[5].position.y) / 2;
                              smooth[3].position.y := smoothHeight;
                           end;
                         end;

                         if tree^.neighbour[1] <> NIL then begin
                           if tree^.neighbour[1].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[1].vertexData[6].position.y + tree^.neighbour[1].vertexData[7].position.y) / 2;
                              smooth[1].position.y := smoothHeight;
                           end;
                         end;
                    end;

                 1: begin
                         if tree^.neighbour[1] <> NIL then begin
                           if tree^.neighbour[1].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[1].vertexData[7].position.y + tree^.neighbour[1].vertexData[8].position.y) / 2;
                              smooth[1].position.y := smoothHeight;
                           end;
                         end;

                         if tree^.neighbour[2] <> NIL then begin
                           if tree^.neighbour[2].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[2].vertexData[0].position.y + tree^.neighbour[2].vertexData[3].position.y) / 2;
                              smooth[5].position.y := smoothHeight;
                           end;
                         end;
                    end;

                 2: begin
                         if tree^.neighbour[3] <> NIL then begin
                           if tree^.neighbour[3].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[3].vertexData[1].position.y + tree^.neighbour[3].vertexData[2].position.y) / 2;
                              smooth[7].position.y := smoothHeight;
                           end;
                         end;

                         if tree^.neighbour[2] <> NIL then begin
                           if tree^.neighbour[2].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[2].vertexData[3].position.y + tree^.neighbour[2].vertexData[6].position.y) / 2;
                              smooth[5].position.y := smoothHeight;
                           end;
                         end;
                    end;

                 3: begin
                         if tree^.neighbour[3] <> NIL then begin
                           if tree^.neighbour[3].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[3].vertexData[0].position.y + tree^.neighbour[3].vertexData[1].position.y) / 2;
                              smooth[7].position.y := smoothHeight;
                           end;
                         end;

                         if tree^.neighbour[0] <> NIL then begin
                           if tree^.neighbour[0].render then begin
                              if not copied then begin
                                for i := 0 to 8 do begin
                                    smooth[i] := tree^.vertexData[i];
                                end;
                                copied := true;
                              end;
                              smoothHeight := (tree^.neighbour[0].vertexData[8].position.y + tree^.neighbour[0].vertexData[5].position.y) / 2;
                              smooth[3].position.y := smoothHeight;
                           end;
                         end;
                    end;
            end;


            m_pd3dDevice.SetTexture(0, tree^.texture);
{    m_pd3dDevice.SetTexture(1, detail_texture);
    m_pd3dDevice.SetTexture(2, lightmap);}
//            m_pd3dDevice.SetTexture(0, overall_textures[1]);

            if copied then
               m_pd3dDevice.DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, 9, 8, vertexIndices, D3DFMT_INDEX16, smooth, sizeOf(TCustomVertex))
            else
               m_pd3dDevice.DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, 9, 8, vertexIndices, D3DFMT_INDEX16, tree^.vertexData, sizeOf(TCustomVertex));

            inc(polygonCount, 8);

         end else begin
             if tree^.hasChildren then begin
                for i := 0 to 3 do begin
                    renderQuadTree(tree^.children[i], testChildren);
                end;
             end;
         end;
     end;
end;

procedure CLandscape.fillNeighbours;
var
   x, y, i: integer;
begin
     for y := tree^.top to ((tree^.top + tree^.size * 2) - 1) do begin
       for x := tree^.left to ((tree^.left + tree^.size * 2) - 1) do begin
           neighbours[x, y, tree^.level].tree := tree;
           neighbours[x, y, tree^.level].quadrant := tree^.quadrant;
       end;
     end;

     if tree^.hasChildren then begin
       for i := 0 to 3 do begin
           fillNeighbours(tree^.children[i]);
       end;
     end;
end;

procedure CLandscape.assignNeighbours;
var
   x, y, i: integer;
begin
// Quadrant:
//
//  0     1
//   -----
//   |   |
//   -----
//  3     2

//  Neighbour
//
//     1
//   -----
// 0 |   |  2
//   -----
//     3

     if tree^.level <= 0 then begin

        tree^.neighbour[0] := NIL;
        tree^.neighbour[1] := NIL;
        tree^.neighbour[2] := NIL;
        tree^.neighbour[3] := NIL;

     end else begin

       case tree^.quadrant of
            0: begin
                    x := tree^.left - 1;
                    y := tree^.top;

                    if (x >= 0) and (y >= 0) then begin
                       tree^.neighbour[0] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[0] := NIL;
                    end;

                    x := tree^.left;
                    y := tree^.top - 1;

                    if (x >= 0) and (y >= 0) then begin
                       tree^.neighbour[1] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[1] := NIL;
                    end;

                    tree^.neighbour[2] := NIL;
                    tree^.neighbour[3] := NIL;
               end;

            1: begin
                    x := tree^.left;
                    y := tree^.top - 1;

                    if (x >= 0) and (y >= 0) then begin
                       tree^.neighbour[1] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[1] := NIL;
                    end;

                    x := tree^.left + (tree^.size * 2) + 1;
                    y := tree^.top;

                    if (x < mapSize) and (y >= 0) then begin
                       tree^.neighbour[2] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[2] := NIL;
                    end;

                    tree^.neighbour[0] := NIL;
                    tree^.neighbour[3] := NIL;
               end;

            2: begin
                    x := tree^.left;
                    y := tree^.top + (tree^.size * 2) + 1;

                    if (x >= 0) and (y < mapSize) then begin
                       tree^.neighbour[3] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[3] := NIL;
                    end;

                    x := tree^.left + (tree^.size * 2) + 1;
                    y := tree^.top;

                    if (x < mapSize) and (y >= 0) then begin
                       tree^.neighbour[2] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[2] := NIL;
                    end;

                    tree^.neighbour[0] := NIL;
                    tree^.neighbour[1] := NIL;
               end;

            3: begin
                    x := tree^.left;
                    y := tree^.top + (tree^.size * 2) + 1;

                    if (x >= 0) and (y < mapSize) then begin
                       tree^.neighbour[3] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[3] := NIL;
                    end;

                    x := tree^.left - 1;
                    y := tree^.top;

                    if (x >= 0) and (y >= 0) then begin
                       tree^.neighbour[0] := neighbours[x, y, tree^.level - 1].tree;
                    end else begin
                       tree^.neighbour[0] := NIL;
                    end;

                    tree^.neighbour[1] := NIL;
                    tree^.neighbour[2] := NIL;
               end;
       end;

     end;

     if tree^.hasChildren then begin
        for i := 0 to 3 do begin
            assignNeighbours(tree^.children[i]);
        end;
     end;

end;


procedure CLandscape.createLightMap;
type
    TColor = packed record
           b:byte;
           g:byte;
           r:byte;
           a:byte;
    end;

var
   x, y, x_ofs, y_ofs: integer;
   zeiger: pointer;
   lockedRect: TD3DLOCKED_RECT;
   pixel:TColor;
   lightAmount: single;
   r, g, b: integer;
begin
     if D3DXCreateTexture(m_pd3dDevice, mapSize, mapSize, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, lightmap) <> D3D_OK then exit;

     if FAILED(lightmap.LockRect(0, lockedRect, nil, 0)) then exit;
     y_ofs := 0;
     for y := 0 to mapSize-1 do begin

         x_ofs := 0;

         for x := 0 to mapSize-1 do begin

             lightAmount := D3DXVec3Dot(normals[x, y].average, sun.direction);
             if (lightAmount < 0) then lightAmount := 0;
             r := trunc((sun.color.r * lightAmount + sun.ambient.r) * 255);
             g := trunc((sun.color.g * lightAmount + sun.ambient.g) * 255);
             b := trunc((sun.color.b * lightAmount + sun.ambient.b) * 255);

             if (r > 255) then r := 255;
             if (g > 255) then g := 255;
             if (b > 255) then b := 255;

             pixel.r := r;
             pixel.g := g;
             pixel.b := b;
             pixel.a := 0;

             zeiger := ptr(integer(lockedRect.pBits) + y_ofs + x_ofs);
             move(pixel, zeiger^, SizeOf(TColor));

             inc(x_ofs, SizeOf(TColor));

         end;
         inc(y_ofs, lockedRect.Pitch);

     end;
     lightmap.UnlockRect(0); 
end;

end.
